//
// Copyright Johan Fabry, 2009
// Licensed under a Creative Commons Attribution-Noncommercial-Share Alike 2.0 Chile license
//
// This class (the core of the plugin) is the result of lots of puzzling on my part,
// with help from Andrew Eisenberg and Andy Clement.
// I share it under the above license so that if you find it useful and incorporate it,
// people using your code can get the same benefits that you got from this work,
// and under the same conditions that you did. That's only fair, right? 
//
// Please be kind and satisfy my basic curiosity by telling me if you are using or reusing this code.

package cl.pleiad.xcutreport;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.aspectj.ajde.core.AjCompiler;
import org.aspectj.ajdt.internal.core.builder.AjState;
import org.aspectj.ajdt.internal.core.builder.IncrementalStateManager;
import org.aspectj.asm.AsmManager;
import org.aspectj.asm.IProgramElement;
import org.aspectj.asm.IRelationship;
import org.aspectj.asm.IRelationship.Kind;
import org.aspectj.asm.internal.Relationship;
import org.aspectj.bridge.ISourceLocation;
import org.aspectj.weaver.World;
import org.aspectj.weaver.patterns.Declare;
import org.aspectj.weaver.patterns.DeclarePrecedence;
import org.aspectj.weaver.patterns.TypePattern;
import org.eclipse.ajdt.core.AspectJPlugin;
import org.eclipse.ajdt.core.javaelements.AJCodeElement;
import org.eclipse.ajdt.core.javaelements.AJCompilationUnit;
import org.eclipse.ajdt.core.javaelements.AdviceElement;
import org.eclipse.ajdt.core.javaelements.AspectElement;
import org.eclipse.ajdt.core.javaelements.AspectJMemberElement;
import org.eclipse.ajdt.core.javaelements.DeclareElement;
import org.eclipse.ajdt.core.javaelements.IntertypeElement;
import org.eclipse.ajdt.core.javaelements.PointcutElement;
import org.eclipse.ajdt.core.lazystart.IAdviceChangedListener;
import org.eclipse.ajdt.core.model.AJProjectModelFacade;
import org.eclipse.ajdt.core.model.AJProjectModelFactory;
import org.eclipse.ajdt.core.model.AJRelationshipManager;
import org.eclipse.ajdt.core.model.AJRelationshipType;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.core.SourceMethod;

public class AdviceListener implements IAdviceChangedListener{

	//This plugin is a singleton so there are no possible conflicts here
	public static ExtendedReportWriter extendedRW = new ExtendedReportWriter("XCutReportOutput");

	private static AdviceListener instance = new AdviceListener();
	
	public static AdviceListener getInstance() { return instance;}

	public void adviceChanged() {

	    extendedRW.clear();
	    Vector<IProject> projectsVector = new Vector<IProject>();
		
	    extendedRW.writeLine("Projects list");
	    
		IProject[] projects = ResourcesPlugin.getWorkspace().getRoot().getProjects();
		for(IProject project : projects){
		     // the AJProjectModelFacade is a facade into the AspectJ's model.  Provides functionality
		     // to query the AspectJ cross cutting model as well as to translate back and forth between
		     // JDT elements and AspectJ elements
		     AJProjectModelFacade model = AJProjectModelFactory.getInstance().getModelForProject(project);
		     
		     // check first to see if there is a model
		     // will return false if there has not yet been successful build of this project
		     if (model.hasModel()) {
		    	 projectsVector.add(project);
		    	 extendedRW.writeLine(project.getName());
		     }
		}
	    extendedRW.newLine();
		
		for (IProject project : projectsVector){
		     AJProjectModelFacade model = AJProjectModelFactory.getInstance().getModelForProject(project);
	    	 extendedRW.writeLine("Project");
	    	 extendedRW.writeLine(project.getName()+"\n");
	    	 reportPrecedence(project, model);
	    	 reportAspects(project, model);
	    	 reportAdvisedBy(model);
	    	 /* Experimental Intertype Declaration Reporting -- off for the moment */
	    	 //reportITD(project,model);
	    	 extendedRW.newLine();
		}
		extendedRW.write("EOF");
		extendedRW.endReport();
	}


	public Set<IProgramElement> findElementsOfKind(IProgramElement container, IProgramElement.Kind kind){
		HashSet<IProgramElement> retval = new HashSet<IProgramElement>();

		if(container.getKind() == kind)
			retval.add(container);
		
		for(Object child : container.getChildren()){
			if (child instanceof IProgramElement){
				Set<IProgramElement> grandchildren = findElementsOfKind((IProgramElement)child,kind);
				for(IProgramElement gc : grandchildren)
					retval.add(gc);
			}
		}
		return retval;
	}
	
	private void reportAspects(IProject project, AJProjectModelFacade model){
		
		AjCompiler compiler = AspectJPlugin.getDefault().getCompilerFactory().getCompilerForProject(project);
		AsmManager existingState = compiler.getModel();
		Set<IProgramElement> aspects = findElementsOfKind(existingState.getHierarchy().getRoot(),IProgramElement.Kind.ASPECT);

		HashSet<AspectElement> types = new HashSet<AspectElement>(); 
		for(IProgramElement aspect : aspects){
			types.add((AspectElement)model.programElementToJavaElement(aspect));
		}
		
		reportAspectStructure(model, types);
		reportPointcuts(model, types);
		reportAdvice(model, types);
	}

	private void reportAspectStructure(AJProjectModelFacade model, HashSet<AspectElement> types) {
		extendedRW.writeLine("Aspect Structure\nAspect Name\tSuperAspect Name (or null)\tNumber of Advice\tNumber of Named Pointcuts\tLines Of Code\tAbstract or Final\tAccessibility(Public or Package)\tFilePath\tEOF");
		for(AspectElement ae : types){
			try {
				extendedRW.write(ae.getFullyQualifiedName()+"\t");
				ae.getSuperclassName(); //sometimes it does not work on the first go ... I love Eclipse ... Sometimes
				extendedRW.write(ae.getSuperclassName() + "\t");
				extendedRW.write(ae.getAdvice().length+"\t");
				extendedRW.write(ae.getPointcuts().length+"\t");
				ISourceLocation loc = model.javaElementToProgramElement(ae).getSourceLocation();
				extendedRW.write(loc.getEndLine()-loc.getLine()+"\t");
				
				List modifs = ae.getAJModifiers();
				if (modifs.isEmpty())
					extendedRW.write("-\t");
				else
					extendedRW.write(modifs.get(0).toString()+"\t");
				
				extendedRW.write(ae.getAJAccessibility()+"\t");
				extendedRW.write(ae.getCompilationUnit().getPath().toString()+"\t");
				extendedRW.writeLine("EOF");
			} catch (JavaModelException e) {
				extendedRW.writeLine("***********ERROR GETTING SUPERCLASS NAME for "+ ae + "***********");
			}			
		}
		
		extendedRW.newLine();
	}

	private void reportPointcuts(AJProjectModelFacade model,
			HashSet<AspectElement> types) {
		extendedRW.writeLine("Pointcuts\nAspect Name\tPointcut Start Line Number\tPointcut Name\tPointcut on one line\tAbstract or Final\tAccessibility\tEOF");
		for(AspectElement ae : types){
			try {
				AJCompilationUnit cu = (AJCompilationUnit)ae.getAncestor(IJavaElement.COMPILATION_UNIT);
				cu.requestOriginalContentMode();
				for(PointcutElement pointcut : ae.getPointcuts()){
					extendedRW.write(ae.getFullyQualifiedName()+"\t");
					extendedRW.write(model.getJavaElementLineNumber(pointcut)+"\t");
					extendedRW.write(pointcut.getElementName()+"\t");
					ISourceRange range = pointcut.getSourceRange();
					extendedRW.write(cu.getSource().substring(range.getOffset(),range.getOffset()+range.getLength()).replaceAll("\\s+", " ")+"\t");
					
					List modifs = pointcut.getAJModifiers();
					if (modifs.isEmpty())
						extendedRW.write("-\t");
					else
						extendedRW.write(modifs.get(0).toString()+"\t");
					
					extendedRW.write(pointcut.getAJAccessibility().toString()+"\t");
					extendedRW.writeLine("EOF");
				}
				cu.discardOriginalContentMode();
			}
			catch (JavaModelException e){
				extendedRW.writeLine("***********ERROR GETTING FQNAME for "+ ae + "***********");
			}
		}

		extendedRW.newLine();
	}

	private void reportAdvice(AJProjectModelFacade model,
			HashSet<AspectElement> types) {
		extendedRW.writeLine("Advice\nAspect Name\tAdvice Start Line Number\tAdvice End Line Number\tAdvice Type\tAdvice Signature\tPointcut of Advice\tEOF");
		for(AspectElement ae : types){
			try {
				for(AdviceElement advice : ae.getAdvice()){
					extendedRW.write(ae.getFullyQualifiedName()+"\t");
					extendedRW.write(model.getJavaElementLineNumber(advice)+"\t");
					extendedRW.write(model.javaElementToProgramElement(advice).getSourceLocation().getEndLine()+"\t");
					extendedRW.write(advice.getElementName()+"\t");
					extendedRW.write(this.adviceSignatureTrim(this.retrieveSignature(advice))+"\t");
					extendedRW.writeLine("EOF");
					
				}
			}
			catch (JavaModelException e){
				extendedRW.writeLine("***********ERROR GETTING FQNAME for "+ ae + "***********");
			}
		}

		extendedRW.newLine();
	}

	@SuppressWarnings("restriction")
	private void reportPrecedence(IProject project, AJProjectModelFacade model){
		List<IRelationship> precedences = new ArrayList<IRelationship>();
		
		//From AJWorldFacade
		AjCompiler compiler = AspectJPlugin.getDefault().getCompilerFactory().getCompilerForProject(project);
		AjState state = IncrementalStateManager.retrieveStateFor(compiler.getId());
		if (state != null) {
			World world = state.getAjBuildManager().getWorld();
			//End AJWorldFacade

			//I'm only interested in the precedence relationships that are explicitly declared
			List<Declare> decprecs = world.getCrosscuttingMembersSet().getDeclareDominates();
			for (Declare decprec : decprecs){
				TypePattern[] precpats = ((DeclarePrecedence) decprec).getPatterns().getTypePatterns();

				//a declare precedence with only one pattern does not make sense, but may happen
				if(precpats.length >= 2){
					for(int i=0; i<precpats.length-1; i++){
						//I'm assuming it's an array of ExactTypePatterns
						String source = precpats[i].getExactType().getBaseName();
						ArrayList<String> dests = new ArrayList<String>(1);
						dests.add(precpats[i+1].getExactType().getBaseName());

						//source and dests are actually not in the right form 
						//for AJProjectModelFacade>>programElementToJavaElement to work
						//but I dont know how to get that form, and I dont mind it like this.
						IRelationship precedence = new Relationship("precedence",Kind.DECLARE,source,dests,false);
						precedences.add(precedence);
					}
				}
			}
		}
		extendedRW.writeLine("Precedence\nSource\tDest\tEND");
		for (IRelationship rel:precedences){
			for (Object target : rel.getTargets())
				extendedRW.writeLine(rel.getSourceHandle()+"\t"+target+"\tEND");
		}
		extendedRW.newLine();
	}
	
	@SuppressWarnings("restriction")
	private void reportAdvisedBy(AJProjectModelFacade model) {
		AJRelationshipType[] relsTypes = {AJRelationshipManager.ADVISED_BY};
		List<IRelationship> rels = (List<IRelationship>) model.getRelationshipsForProject(relsTypes);

		extendedRW.writeLine("Shadows\nBase FQN\tBase Line Number\tType of Pointcut\tAspect FQN\tPointcut with Test\tAdvice Line Number\tBase Call String\tCall target method\tEND");
		for (IRelationship rel : rels) {
			try {
				IJavaElement source = model.programElementToJavaElement(rel.getSourceHandle());
				ICompilationUnit sourceCU;

				//Check that source is not an aspect (yes, this happens, dont want to find out the semantics of this)
				try {
					sourceCU = (ICompilationUnit)source.getAncestor(IJavaElement.COMPILATION_UNIT);
					String extension = ((IFile)sourceCU.getCorrespondingResource()).getFileExtension();
					if (extension.equals("aj")) continue;
				} catch (JavaModelException e1) {
					extendedRW.writeLine("***********ERROR GETTING SOURCE INFO for "+ source + "***********");
					continue;
				}

				for (String targetHandle : (Iterable<String>) rel.getTargets()) {
					String baseCode, calltarget;
					//Type of advice is execution
					if (source instanceof SourceMethod){
						SourceMethod match = (SourceMethod)source;
						//Fully Qualified Class name of base code
						extendedRW.write(match.getDeclaringType().getFullyQualifiedName('.')+"\t");
						ISourceRange sourceRange = match.getSourceRange();
						extendedRW.write(""+model.getJavaElementLineNumber(match)+"\t");
						extendedRW.write("execution\t");
						baseCode = calltarget = "-";
					}
					//type of advice is call
					else if(source instanceof AJCodeElement) {
						//Fully Qualified Class name of base code
						AJCodeElement match = (AJCodeElement)source;
						extendedRW.write(((IType)match.getAncestor(IJavaElement.TYPE)).getFullyQualifiedName('.')+"\t");
						ISourceRange sourceRange = match.getNameRange();
						extendedRW.write(""+model.getJavaElementLineNumber(match)+"\t");
						extendedRW.write("call\t");
						String baseCodeCU = sourceCU.getSource();
						baseCode = retrieveCallExpression(sourceRange, baseCodeCU);
						calltarget = extractCallTarget(match);
					}
					else {
						extendedRW.writeLine("***********ERROR GETTING SOURCE INFO: unknown type of source for: "+ source + "***********");
						continue;
					}

					IJavaElement target = model.programElementToJavaElement(targetHandle);
					//Fully Qualified Name of aspect
					extendedRW.write(((IType)target.getAncestor(IJavaElement.TYPE)).getFullyQualifiedName('.')+"\t");
					//Runtime test 
					extendedRW.write(rel.hasRuntimeTest()+"\t");					
					extendedRW.write(""+model.getJavaElementLineNumber(target)+"\t");
					extendedRW.write(baseCode+"\t");
					extendedRW.write(calltarget+"\t");
					extendedRW.writeLine("END");
				}
			}
			catch(Exception e){
				extendedRW.writeLine("*********** Unhandled Exception while processing relationship: "+e.toString());
			}
		}
		extendedRW.newLine();
	}
	
	/* EXPERIMENTAL intertype declaration reporting */
	private void reportITD(IProject project, AJProjectModelFacade model) {
		AJRelationshipType[] relsTypes = {AJRelationshipManager.DECLARED_ON};
		List<IRelationship> rels = (List<IRelationship>) model.getRelationshipsForProject(relsTypes);
		
		extendedRW.writeLine("ITD\nBase FQN\tAspect FQN\tMethod Start Line Number\tMethod End Line Number\tMethod Signature\tEND");

		for (IRelationship rel : rels) {
			try{
			if (rel.getKind() == Kind.DECLARE_INTER_TYPE){
					Object ssrc = model.programElementToJavaElement(rel.getSourceHandle());
					if(ssrc instanceof DeclareElement) continue;
					IntertypeElement source = (IntertypeElement)ssrc;
				IJavaElement target = model.programElementToJavaElement((String)rel.getTargets().get(0));
				extendedRW.write(((org.eclipse.jdt.internal.core.SourceType)target).getFullyQualifiedName()+"\t");
				extendedRW.write(((IType)source.getAncestor(IJavaElement.TYPE)).getFullyQualifiedName('.')+"\t");
				extendedRW.write(""+model.javaElementToProgramElement(source).getSourceLocation().getLine()+"\t");
				extendedRW.write(""+model.javaElementToProgramElement(source).getSourceLocation().getEndLine()+"\t");
				try {
					String src = this.retrieveSignature(source);
					int idx = src.indexOf("{");
					//TODO check this!
					if(idx == -1)// a field ??
						extendedRW.writeLine(src+"\t");
					else
						extendedRW.writeLine(src.substring(0, idx)+"\t");
					/*
					extendedRW.write(source.getReturnType()+" "+source.getElementName()+"(");
					IntertypeElementInfo elem;
					elem = (IntertypeElementInfo)source.getElementInfo();
					String[] args = source.getParameterTypes();
					if (args != null)
						for(String name : args)
							extendedRW.write(name+" ");
					extendedRW.writeLine(")");*/
				} catch (JavaModelException e) {
					extendedRW.writeLine("***********Error getting element info of: "+ source + "***********");
				}
			}
		}
		catch(Exception ex){
			System.out.println("***********Error while reporting ITD ***********");}
		}
	}
	
	private String extractCallTarget(AJCodeElement match) {
		String matchstring = match.getName();
		int start = matchstring.indexOf('(')+1;
		return matchstring.substring(start, matchstring.length()-1);
	}

	private String retrieveCallExpression(ISourceRange sourceRange, String baseCode) {
		return baseCode.substring(sourceRange.getOffset(), sourceRange.getOffset()+sourceRange.getLength()).replaceAll("\\s+", " ");
	}

	private String retrieveSignature(AspectJMemberElement advice) throws JavaModelException{
		ISourceRange range = advice.getSourceRange();
		AJCompilationUnit cu = (AJCompilationUnit)advice.getAncestor(IJavaElement.COMPILATION_UNIT);
		ISourceRange nameRange = advice.getNameRange();
		cu.requestOriginalContentMode();
		String source = cu.getSource().substring(nameRange.getOffset(), range.getOffset()+range.getLength());
		cu.discardOriginalContentMode();
		return source;
	}
	
	private String adviceSignatureTrim(String source){
		int dp = source.indexOf(':');
		int max = source.indexOf('{');
		
		//compress the sig into one line 
		String sig = source.substring(dp+1,max).replaceAll("\\s+", " ");
		return source.substring(0,dp).trim()+"\t"+sig.trim();
	}

}
