package cl.pleiad.xcutreport;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import org.eclipse.ui.*;
import org.eclipse.ui.console.*;

import cl.pleiad.xcutreport.actions.ToolbarAction;

public class ExtendedReportWriter extends ReportWriter {
	public String fileName;
	public static boolean bConsoleOutput, bFileOutput;

  	public ExtendedReportWriter(final String fileName) {
  		// Default values:
  		bConsoleOutput = false;
  		bFileOutput = false;
  		this.fileName = fileName + ".xcr";
  	}

	public void unsetFileOutput() {	bFileOutput = false;	}

	public void toggleConsoleOutput() {	bConsoleOutput = !bConsoleOutput;	}
	public void toggleFileOutput() {	bFileOutput = !bFileOutput;	}

	public boolean consoleOutput() {	return bConsoleOutput;	}
	public boolean fileOutput() {	return bFileOutput;	}

  	public String getFileName() { return fileName; }
  	public void setFileName(String newName) {
  		if( newName != null )
  			fileName = newName;
  	}

  	private void writeToFile() {
  		try {
  			FileOutputStream fout = new FileOutputStream(fileName);
  			new PrintStream(fout).print(output);
  			fout.close();
  		}
  		catch (IOException e) {
  			System.err.println("Unable to write to file " + fileName);
  			e.printStackTrace();
  		}
  	}

	private void writeToConsole() {
		try {
            final MessageConsole myConsole = new MessageConsole("XCut Report Console", null);
			ConsolePlugin.getDefault().getConsoleManager().addConsoles(new IConsole[]{myConsole});                     

			MessageConsoleStream out = myConsole.newMessageStream();
			out.print(output.toString());

			IWorkbenchPage page = ToolbarAction.window.getActivePage();

			String id = IConsoleConstants.ID_CONSOLE_VIEW;
			IConsoleView view = (IConsoleView) page.showView(id);
			view.display(myConsole);
		} catch (IllegalStateException ise) {
			System.err.println("Illegal state exception.");
			ise.printStackTrace();
		} catch (PartInitException e) {
			System.err.println("A workbench part cannot be initialized correctly.");
			e.printStackTrace();
		}
	}

	@Override
	public void endReport() {
		if( bConsoleOutput )	writeToConsole();
		if( bFileOutput )	writeToFile();
	}
}
